from collections import defaultdict
from copy import deepcopy


def get_producer_config(config):
    config = deepcopy(config)
    producer_config = config.get('producer')
    producer_config.update({
        'bootstrap.servers': config['bootstrap.servers'],
        'schema.registry.url': config['schema.registry.url']
    })
    return producer_config


def get_loader_config(config):
    config = deepcopy(config)
    loader_config = defaultdict(dict, config.get('loader'))
    loader_config.update({
        'bootstrap.servers': config['bootstrap.servers'],
        'schema.registry.url': config['schema.registry.url']
    })
    loader_config['consumer'].update({
        'bootstrap.servers': config['bootstrap.servers'],
        'schema.registry.url': config['schema.registry.url']
    })
    return loader_config



import structlog
from redis import StrictRedis
from redis.sentinel import Sentinel

from confluent_kafka_helpers.message import Message

from eventsourcing_helpers.messagebus.backends.kafka.offset_watchdog.backends import (  # noqa
    OffsetWatchdogBackend
)

logger = structlog.get_logger(__name__)

__all__ = ['RedisOffsetWatchdogBackend']

SOCKET_TIMEOUT = 0.1
SOCKET_CONNECT_TIMEOUT = 0.1


class RedisOffsetWatchdogBackend(OffsetWatchdogBackend):
    """
    Redis offset watchdog backend.
    Stores the last offsets in a Redis database.
    """

    def __init__(self, config: dict) -> None:
        super().__init__(config=config)
        self._redis = self._redis_sentinel = None
        socket_connect_timeout = config.get(
            'socket_connect_timeout', SOCKET_CONNECT_TIMEOUT
        )
        socket_timeout = config.get('socket_timeout', SOCKET_TIMEOUT)
        if 'redis_uri' in config:
            assert 'redis_sentinels' not in config
            self._redis = StrictRedis.from_url(
                url=config['redis_uri'],
                socket_connect_timeout=socket_connect_timeout,
                socket_timeout=socket_timeout, retry_on_timeout=True,
                decode_responses=True
            )

        else:
            assert 'redis_sentinels' in config
            assert 'redis_sentinel_service_name' in config

            sentinels = [
                tuple(h.split(':'))
                for h in config['redis_sentinels'].split(',')
            ]

            self._redis_sentinel = Sentinel(
                sentinels, db=config['redis_database'],
                socket_connect_timeout=socket_connect_timeout,
                socket_timeout=socket_timeout, retry_on_timeout=True,
                decode_responses=True
            )
            self._redis_sentinel_service_name = config[
                'redis_sentinel_service_name'
            ]

    @property
    def redis(self) -> StrictRedis:
        if self._redis:
            return self._redis
        return self._redis_sentinel.master_for(
            self._redis_sentinel_service_name
        )

    def seen(self, message: Message) -> bool:
        last_offset = self.redis.get(self._key(message))
        if last_offset is None:
            return False

        # https://github.com/edenhill/librdkafka/issues/1720
        # until this bug is fixed at least make sure we log when it happens.
        offset = message._meta.offset
        offset_diff = offset - int(last_offset)
        if offset_diff != 1:
            logger.warning(
                "Offset deviation detected", offset_diff=offset_diff
            )

        return offset <= int(last_offset)

    def set_seen(self, message: Message):
        self.redis.set(self._key(message), message._meta.offset)



from typing import Any, Callable, Generator

from eventsourcing_helpers.models import Entity
from eventsourcing_helpers.repository import Repository
from eventsourcing_helpers.serializers import from_message_to_dto


class ESEntityBuilder:
    """
    This class is responsible for loading the messages and
    return the aggregate in the requested state.

    >> builder = ESEntityBuilder(repository_config=config)
    >> builder.rebuild(entity_class=Order, id="123131323132")
    Order(id="123131323132", state="completed", ...)
    """

    def __init__(
        self, repository_config: dict,
        message_deserializer: Callable = from_message_to_dto,
        repository: Any = Repository
    ) -> None:
        self.message_deserializer = message_deserializer
        self.repository = repository(repository_config)

    def rebuild(self, entity_class: Entity, id: str, max_offset=None) -> Entity:
        """
        Rebuild the entity state based on the events on the repository.

        Args:
            entity_class: Class used to apply the events
            id: entity id used to load the events
            max_offset: stop to read at this offset

        Returns:
            entity: rebuilt entity
        """
        entity = entity_class()
        entity._apply_events(self._get_events(id, max_offset=max_offset),
                             ignore_missing_apply_methods=True)
        return entity

    def _get_events(self, id: str,
                    max_offset: int = None) -> Generator[Any, None, None]:
        """
        Get events for the given id and stop in the max offset.

        Args:
            id: id used to load events from repository
            max_offset: when given is used to stop the reading of the messages
        """
        with self.repository.load(id) as events:
            for event in events:
                if max_offset is not None and event._meta.offset > max_offset:
                    break
                yield self.message_deserializer(event)



from typing import Any, Callable, Generator

from confluent_kafka_helpers.loader import AvroMessageLoader
from confluent_kafka_helpers.producer import AvroProducer

from eventsourcing_helpers.repository.backends import RepositoryBackend
from eventsourcing_helpers.repository.backends.kafka.config import (
    get_loader_config, get_producer_config
)
from eventsourcing_helpers.serializers import to_message_from_dto


class KafkaAvroBackend(RepositoryBackend):
    def __init__(
        self, config: dict, producer=AvroProducer, loader=AvroMessageLoader,
        value_serializer: Callable = to_message_from_dto,
        get_producer_config: Callable = get_producer_config,
        get_loader_config: Callable = get_loader_config
    ) -> None:
        producer_config = get_producer_config(config)
        loader_config = get_loader_config(config)

        self.producer, self.loader = None, None
        if producer_config:
            self.producer = producer(
                producer_config, value_serializer=value_serializer
            )
        if loader_config:
            self.loader = loader(loader_config)

    def commit(self, id: str, events: list, **kwargs) -> None:
        """
        Commit events to Kafka.

        Args:
            id: Aggregate root id to be used as key in the message.
            events: List of staged events to be committed.
        """
        assert self.producer is not None, "Producer is not configured"

        # TODO: investigate how "exactly once" delivery works
        # right now there is a potential risk that one of the produce's
        # fails and leaving the aggregate in an invalid state.
        for event in events:
            self.producer.produce(key=id, value=event, **kwargs)

    def load(self, id: str, **kwargs) -> list:
        """
        Load events from Kafka.

        Args:
            id: Aggregate root id to load.

        Returns:
            list: Loaded events.
        """
        assert self.loader is not None, "Loader is not configured"
        return self.loader.load(id, **kwargs)

    def get_events(
        self, id: str, message_deserializer: Callable
    ) -> Generator[Any, None, None]:
        """
        Get all aggregate events from the event storage.

        Args:
            id: ID of the aggregate root.

        Returns:
            list: List with all events.
        """
        with self.load(id) as events:  # type:ignore
            for event in events:
                yield message_deserializer(event, is_new=False)



from collections import defaultdict
from copy import deepcopy


def get_producer_config(config):
    config = deepcopy(config)
    producer_config = config.get('producer', None)
    if not producer_config:
        return None

    producer_config.update(
        {
            'bootstrap.servers': config['bootstrap.servers'],
            'schema.registry.url': config['schema.registry.url']
        }
    )
    return producer_config


def get_consumer_config(config):
    config = deepcopy(config)
    consumer_config = config.get('consumer')
    if not consumer_config:
        return None

    consumer_config.pop('offset_watchdog', None)
    consumer_config.update(
        {
            'bootstrap.servers': config['bootstrap.servers'],
            'schema.registry.url': config['schema.registry.url']
        }
    )
    return consumer_config


def get_offset_watchdog_config(config):
    config = deepcopy(config)
    consumer_config = config.get('consumer', {})
    if not consumer_config:
        return None

    offset_wd_config = defaultdict(
        dict, consumer_config.get('offset_watchdog', {})
    )
    offset_wd_config['backend_config'].update(
        {'group.id': consumer_config['group.id']}
    )
    return offset_wd_config



from unittest.mock import MagicMock

from eventsourcing_helpers.builders import ESEntityBuilder
from eventsourcing_helpers.models import Entity


class TestEntity(Entity):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.state = None

    def apply_first_event(self, event):
        self.state = "first_event"

    def apply_second_event(self, event):
        self.state = "second_event"

    def apply_third_event(self, event):
        self.state = "third_event"


class ESEntityBuilderTests:

    def setup_method(self):
        events = [
            MagicMock(id="123", _class="FirstEvent", _meta=MagicMock(offset=0)),  # noqa
            MagicMock(id="123", _class="SecondEvent", _meta=MagicMock(offset=1)), # noqa
            MagicMock(id="123", _class="ThirdEvent", _meta=MagicMock(offset=2))
        ]

        config = {
            'return_value.load.return_value.__enter__.return_value': events
        }
        repository = MagicMock()
        repository.configure_mock(**config)

        self.factory = ESEntityBuilder(
            repository_config={},
            message_deserializer=lambda e: e,
            repository=repository,
        )

    def test_get_events_stop_with_max_offset(self):
        events_max_offset1 = list(self.factory._get_events(id="123", max_offset=1))  # noqa
        events_max_offset2 = list(self.factory._get_events(id="123", max_offset=2))  # noqa

        assert len(events_max_offset1) == 2
        assert len(events_max_offset2) == 3

    def test_rebuild_returns_the_entity_to_the_final_state(self):
        entity = self.factory.rebuild(TestEntity, id="123")
        assert entity.state == "third_event"

    def test_rebuild_returns_the_entity_in_the_right_state_when_max_offset_is_given(self):  # noqa
        entity = self.factory.rebuild(TestEntity, id="123", max_offset=0)
        assert entity.state == "first_event"

        entity = self.factory.rebuild(TestEntity, id="123", max_offset=1)
        assert entity.state == "second_event"

        entity = self.factory.rebuild(TestEntity, id="123", max_offset=2)
        assert entity.state == "third_event"
